// FFT of a 2D RGB image computed on the GPU.
//
// Usage: Provide image as command line argument, 
// press <SPACE> to switch between image and FFT.
//
// This demo uses the following software packages:
// GLUT, GLEW, SOIL, gpu_fft, load_shaders, Timer
//
// Code written by Jeppe Revall Frisvad, August 2009.
// Copyright (c) DTU Informatics 2009

#include <iostream>
#include <GL/glew.h>
#include <GL/glut.h>
#include "../SOIL/SOIL.h"
#include "load_shaders.h"
#include "FFT.h"
#include "OSTimer.h"

using namespace std;

//////////////////////////////////////////////////////////////////////
// Globals
//////////////////////////////////////////////////////////////////////

namespace
{
  int winw = 512, winh = 512;
  const unsigned int FFT_SIZE = 512;
  unsigned int source_list;
  unsigned int source_tex;
  unsigned int g2r_prog;
  bool draw_fft = false; 
  FFT* fft;

  GLcharARB* texcoord_vert =
    "void main()\n"
    "{\n"
    "  gl_TexCoord[0] = gl_MultiTexCoord0;\n"
    "  gl_Position = ftransform();\n"
    "}";

  GLcharARB* g2r_frag =
    "uniform sampler2D source;\n"
    "void main()\n"
    "{\n"
    "  float g = texture2D(source, gl_TexCoord[0].xy).g;\n"
    "  gl_FragColor.rgb = vec3(g, 0.0, 0.0);\n"
    "}";
}
  
//////////////////////////////////////////////////////////////////////
// Draw functions
//////////////////////////////////////////////////////////////////////

void set_projection()
{
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(-1.0, 1.0, -1.0, 1.0, -1.0, 1.0);
  glMatrixMode(GL_MODELVIEW);
}

void draw_source()
{
  glClearColor(0.0, 0.0, 0.0, 0.0);
  glClear(GL_COLOR_BUFFER_BIT);
  set_projection();
  glLoadIdentity();

  glColor3f(1.0f, 1.0f, 1.0f);
  glCallList(source_list);
}

void draw_fft_source_rb()
{
  glClearColor(0.0, 0.0, 0.0, 0.0);
  glClear(GL_COLOR_BUFFER_BIT);
  set_projection();
  glLoadIdentity();

  glColor3f(1.0f, 0.0f, 1.0f);
  glCallList(source_list);
}

void draw_fft_source_g()
{
  glClearColor(0.0, 0.0, 0.0, 0.0);
  glClear(GL_COLOR_BUFFER_BIT);
  set_projection();
  glLoadIdentity();

  glUseProgram(g2r_prog);
  glEnable(GL_TEXTURE_2D);
  glCallList(source_list);
  glDisable(GL_TEXTURE_2D);
  glUseProgram(0);
}

//////////////////////////////////////////////////////////////////////
// Initialization
//////////////////////////////////////////////////////////////////////

void GLEWinit()
{
  GLenum err = glewInit();
  if(err != GLEW_OK)
  {
    cerr << "Error loading GLEW: " << glewGetErrorString(err) << endl;
    exit(0);
  }
}

void GLinit()
{
  glEnable(GL_CULL_FACE);
}

void init_source(const char* filename)
{
  int width, height, channels;
  unsigned char* data = SOIL_load_image(filename, &width, &height, &channels, SOIL_LOAD_RGB);
  if(!data)
  {
    cerr << "Error: could not load source image" << endl;
    exit(0);
  }

  source_tex = SOIL_create_OGL_texture(data, width, height, channels, SOIL_CREATE_NEW_ID, SOIL_FLAG_INVERT_Y);
  if(!source_tex)
  {
    cerr << "Error: could not create texture from source image" << endl;
    exit(0);
  }

  source_list = glGenLists(1);
  glNewList(source_list, GL_COMPILE);

  glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
  glEnable(GL_TEXTURE_2D);
  glBindTexture(GL_TEXTURE_2D, source_tex);
  glBegin(GL_POLYGON);
    glTexCoord2f(0.0f,  0.0f);
    glVertex2f( -1.0f, -1.0f);
    glTexCoord2f(1.0f,  0.0f);
    glVertex2f(  1.0f, -1.0f);
    glTexCoord2f(1.0f,  1.0f);
    glVertex2f(  1.0f,  1.0f);
    glTexCoord2f(0.0f,  1.0f);
    glVertex2f( -1.0f,  1.0f);
  glEnd();
  glDisable(GL_TEXTURE_2D);
  glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, 0);

  glEndList();
}

void init(const char* filename)
{
  GLEWinit();
  GLinit();
  init_source(filename);
  load_shaders(texcoord_vert, g2r_frag, g2r_prog);

  fft = new FFT(FFT_SIZE, FFT_SIZE);
}

//////////////////////////////////////////////////////////////////////
// GLUT Callback Functions
//////////////////////////////////////////////////////////////////////

void display()
{
  static const float scale = 1.0f/static_cast<float>(FFT_SIZE);

  glClear(GL_COLOR_BUFFER_BIT);
  glLoadIdentity();

  if(draw_fft)
  {
    Timer t;
    t.start();

    fft->set_input(draw_fft_source_rb);
    fft->do_fft();

    glBlendFunc(GL_ONE, GL_ONE);
    glEnable(GL_BLEND);
    fft->draw_output(scale, 0.0f, 0.0f, 1);
    fft->draw_output(0.0f, 0.0f, scale, 2);
    glDisable(GL_BLEND);
    
    fft->set_input(draw_fft_source_g);
    fft->redraw_input();
    fft->do_fft();

    glEnable(GL_BLEND);
    fft->draw_output(0.0f, scale, 0.0f);
    glDisable(GL_BLEND);

    t.stop();
    cout << "FFT computed in " << t.get_time() << " seconds." << endl;
  }
  else
    draw_source();

  glutSwapBuffers();
}

void reshape(int width, int height)
{
  winw = width;
  winh = height;

  glViewport(0, 0, winw, winh);
}

void keyboard(unsigned char key, int x, int y)
{
  switch(key)
  {
  case ' ':
    draw_fft = !draw_fft;
    glutPostRedisplay();
    break;
  case 27:
    delete fft;
    exit(0);
  }
}

//////////////////////////////////////////////////////////////////////
// Main
//////////////////////////////////////////////////////////////////////

int main(int argc, char** argv)
{
  glutInit(&argc, argv);
  glutInitWindowSize(winw, winh);
  glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH | GLUT_ALPHA);
  glutCreateWindow("GPU FFT demo");
  glutDisplayFunc(display);
  glutReshapeFunc(reshape);
  glutKeyboardFunc(keyboard);

  init(argc > 1 ? argv[1] : "test.jpg");

  glutMainLoop();

  return 0;
}
